"""
Santec Connection Enums.
"""

from enum import Enum

from ..santec_communication_wrapper import CommunicationMethod, GPIBConnectType
from ..santec_wrapper import CommunicationTerminator


class Terminator(Enum):
    """Communication terminators for Santec instruments."""

    CR = CommunicationTerminator.Cr
    LF = CommunicationTerminator.Lf
    CRLF = CommunicationTerminator.CrLf


class GPIBType(Enum):
    """GPIB connection types for Santec instruments."""

    NIVisa = GPIBConnectType.NIVisa
    NI4882 = GPIBConnectType.NI4882
    KeysightVisa = GPIBConnectType.KeysightIO


class ConnectionType(Enum):
    """Connection types for Santec instruments."""

    USB = CommunicationMethod.USB
    GPIB = CommunicationMethod.GPIB
    TCPIP = CommunicationMethod.TCPIP
    DEV = "DAQ"
    NULL = "Unknown"



"""
Base instrument module.
"""

import inspect
from .wrapper import (
    MPM,
    TSL,
    InstrumentWrapper,
    InstrumentExceptionCode,
    to_instrument_exception_code,
)
from ..logger import get_logger


class BaseInstrument:
    """Base class for instruments."""

    def __init__(self):
        self._instrument = None
        self._status = None
        self.logger = get_logger(self._instrument.__class__.__name__)

    def instrument(self, wrapper_type: InstrumentWrapper):
        """Set the instrument wrapper type."""
        if not wrapper_type:
            return None
        return self._instrument

    def _check_restricted_method(self):
        """Check if the method is restricted to certain instrument types."""
        if not isinstance(self._instrument, (TSL, MPM)):
            stack = inspect.stack()
            caller_frame = stack[1].function
            error_string = (
                f"{self._instrument.__class__.__name__} "
                f"is not allowed to use method '{caller_frame}'."
            )
            self.logger.error(error_string)
            raise PermissionError(error_string)

    @property
    def status(self) -> str:
        """Returns the current instrument status string."""
        return self.__status

    @property
    def __status(self) -> str:
        """Returns the current instrument status string."""
        if self._status is not None:
            return self._status.name
        return InstrumentExceptionCode.Unknown.name

    @__status.setter
    def __status(self, value: InstrumentExceptionCode):
        """Set the current instrument status."""
        self.logger.debug(f"Setting instrument status: {value}")
        self._status = value

    def query(self, command: str) -> str:
        """Send a query command to the instrument
        and return the status and response."""
        """This method is restricted to TSL and MPM instruments."""
        self._check_restricted_method()
        command = command.upper()
        self.logger.debug("Query command: %s", command)

        try:
            status, response = self._instrument.Echo(command, "")
            self.__status = to_instrument_exception_code(status)
            self.logger.debug(f"Query Status: {status}. Response: {response}.")
            return response

        except Exception as e:
            error_string = f"Error while querying command {command}: {e}"
            self.logger.error(error_string)
            raise RuntimeError(error_string)

    def write(self, command: str) -> None:
        """Send a write command to the instrument and return the status."""
        """This method is restricted to TSL and MPM instruments."""
        self._check_restricted_method()
        command = command.upper()
        self.logger.debug("Write command: %s", command)

        try:
            status = self._instrument.Write(command)
            self.__status = to_instrument_exception_code(status)
            self.logger.debug(f"Write Status: {status}.")

        except Exception as e:
            error_string = f"Error while writing command {command}: {e}"
            self.logger.error(error_string)
            raise RuntimeError(error_string)

    def read(self) -> str:
        """Read data from the instrument and return the status and response."""
        """This method is restricted to TSL and MPM instruments."""
        self._check_restricted_method()
        self.logger.debug("Read command.")

        try:
            status, response = self._instrument.Read("")
            self.__status = to_instrument_exception_code(status)
            self.logger.debug(f"Read Status: {status}. Response: {response}.")
            return response

        except Exception as e:
            error_string = f"Error while reading instrument: {e}"
            self.logger.error(error_string)
            raise RuntimeError(error_string)

    @property
    def idn(self):
        """Return the identification string of the instrument."""
        idn = self.query("*IDN?")
        self.logger.debug(f"IDN string: {idn}")
        return idn

    @property
    def product_name(self):
        """Return the product name of the instrument."""
        product_name = self._instrument.Information.ProductName
        self.logger.debug(f"Product name: {product_name}")
        return product_name

    @property
    def serial_number(self):
        """Return the serial number of the instrument."""
        serial_number = self._instrument.Information.SerialNumber
        self.logger.debug(f"Serial number: {serial_number}")
        return serial_number

    @property
    def firmware_version(self):
        """Return the firmware version of the instrument."""
        firmware_version = self._instrument.Information.FWversion
        self.logger.debug(f"Firmware version: {firmware_version}")
        return firmware_version

    def _get_response(self, function_name, *args):
        """Get a response from the instrument for a given function."""
        error_code, response = getattr(self._instrument, function_name)(*args)
        self.__status = to_instrument_exception_code(error_code)
        return response

    def _get_multiple_responses(self, function_name, response_type_1, response_type_2):
        """Get multiple responses from the instrument for a given function."""
        response_1 = self._init_response(response_type_1)
        response_2 = self._init_response(response_type_2)
        error_code, response_1, response_2 = getattr(self._instrument, function_name)(
            response_1, response_2
        )
        self.__status = to_instrument_exception_code(error_code)
        return response_1, response_2

    def _set_and_get_multiple_responses(
        self, function_name, response_type_1, response_type_2, *args
    ):
        """Set values and get multiple responses
        from the instrument for a given function."""
        response_1 = self._init_response(response_type_1)
        response_2 = self._init_response(response_type_2)
        error_code, response_1, response_2 = getattr(self._instrument, function_name)(
            *args, response_1, response_2
        )
        self.__status = to_instrument_exception_code(error_code)
        return response_1, response_2

    def _get_function(self, function_name, response_type):
        """Get a response from the instrument for a given function."""
        response = self._init_response(response_type)
        return self._get_response(function_name, response)

    def _set_function(self, function_name, *args):
        """Set values on the instrument for a given function."""
        error_code = getattr(self._instrument, function_name)(*args)
        self.__status = to_instrument_exception_code(error_code)

    def _set_and_get_function(self, function_name, *args, response_type=-1):
        """Set values and get a response
        from the instrument for a given function."""
        if response_type != -1:
            response = self._init_response(response_type)
            return self._get_response(function_name, *args, response)
        return self._get_response(function_name, *args)

    @staticmethod
    def _init_response(response_type):
        """Initialize a response based on the provided type."""
        return response_type() if isinstance(response_type, type) else None

    def _get_function_enum(self, function_name, function_enum_name):
        """Get an enum value from the instrument for a given function."""
        self._check_restricted_method()
        enum_value = function_enum_name.value
        error_code, enum_value = getattr(self._instrument, function_name)(enum_value)
        self.__status = to_instrument_exception_code(error_code)
        return function_enum_name.__class__(enum_value)

    def _set_function_enum(self, function_name, function_enum_name):
        """Set an enum value on the instrument for a given function."""
        self._check_restricted_method()
        error_code = getattr(self._instrument, function_name)(function_enum_name.value)
        self.__status = to_instrument_exception_code(error_code)

    def _set_and_get_function_enum(self, function_name, enum_type, *args):
        """Set values and get an enum value
        from the instrument for a given function."""
        self._check_restricted_method()
        error_code, enum_value = getattr(self._instrument, function_name)(*args)
        self.__status = to_instrument_exception_code(error_code)
        return enum_type(enum_value)

    def disconnect(self):
        """Disconnect the instrument."""
        self.logger.info("Disconnecting instrument.")
        error_code = self._instrument.DisConnect()
        self.__status = to_instrument_exception_code(error_code)
        self.logger.info(f"Instrument disconnected. Status: {self.__status}.")



"""
Santec Communication DLL Wrapper.
"""

import Santec.Communication as Comm

# Define the main classes and types for communication
GPIBConnectType = Comm.GPIBConnectType
MainCommunication = Comm.MainCommunication
CommunicationMethod = Comm.CommunicationMethod



"""
TSL Instrument Enums.
"""

from enum import Enum
from ..santec_wrapper import TSL


class LDStatus(Enum):
    """Enum for Laser Diode (LD) Status of the TSL instrument."""

    OFF = TSL.LD_Status.LD_OFF
    ON = TSL.LD_Status.LD_ON


class ShutterStatus(Enum):
    """Enum for Shutter Status of the TSL instrument."""

    OPEN = TSL.Shutter_Status.Shutter_Open
    CLOSE = TSL.Shutter_Status.Shutter_Close


class ScanMode(Enum):
    """Enum for Scan Modes of the TSL instrument."""

    STEPPED_ONE_WAY = TSL.Sweep_Mode.Step_Oneway
    CONTINUOUS_ONE_WAY = TSL.Sweep_Mode.Continuous_Oneway
    STEPPED_TWO_WAY = TSL.Sweep_Mode.Step_Twoway
    CONTINUOUS_TWO_WAY = TSL.Sweep_Mode.Continuous_Twoway


class ScanStatus(Enum):
    """Enum for Scan Status of the TSL instrument."""

    STANDBY = TSL.Sweep_Status.Standby
    RUNNING = TSL.Sweep_Status.Running
    PAUSE = TSL.Sweep_Status.Pausing
    STANDING_BY_TRIGGER = TSL.Sweep_Status.WaitingforTrigger
    PREPARATION_FOR_SWEEP_START = TSL.Sweep_Status.Returning


class TriggerOutputMode(Enum):
    """Enum for Trigger Output Modes of the TSL instrument."""

    NONE = getattr(TSL.Trigger_Output_Mode, "None")
    STOP = TSL.Trigger_Output_Mode.Stop
    START = TSL.Trigger_Output_Mode.Start
    STEP = TSL.Trigger_Output_Mode.Step


class TriggerInputMode(Enum):
    """Enum for Trigger Input Modes of the TSL instrument."""

    DISABLE = TSL.Trigger_Input_Mode.Disable
    ENABLE = TSL.Trigger_Input_Mode.Enable


class ScanStartMode(Enum):
    """Enum for Scan Start Modes of the TSL instrument."""

    NORMAL = TSL.Sweep_Start_Mode.Normal
    WAITING_FOR_TRIGGER = TSL.Sweep_Start_Mode.WaitingforTrigger


class PowerUnit(Enum):
    """Enum for Power Units of the TSL instrument."""

    dBm = TSL.Power_Unit.dBm
    mW = TSL.Power_Unit.mW


class WavelengthUnit(Enum):
    """Enum for Wavelength Units of the TSL instrument."""

    nm = TSL.Wavelength_Unit.nm
    THz = TSL.Wavelength_Unit.THz


class PowerMode(Enum):
    """Enum for Power mode of the TSL instrument."""

    AutoCurrentControl = TSL.Power_Mode.ACC
    AutoPowerControl = TSL.Power_Mode.APC


class TriggerOutputSetting(Enum):
    """Enum for Trigger output source of the TSL instrument."""

    WAVELENGTH = TSL.TriggerOut_Source.Wavelength
    TIME = TSL.TriggerOut_Source.Time


class CoherenceControlStatus(Enum):
    """Enum for Coherence Control Status of the TSL instrument."""

    OFF = TSL.Coh_Status.Coh_OFF
    ON = TSL.Coh_Status.Coh_ON


class GPIBDelimiter(Enum):
    """Enum for GPIB Command Delimiter of the TSL instrument."""

    CR = 0
    LF = 1
    CRLF = 2
    NONE = 3



"""
Instrument Manager module.
"""

from typing import Dict
from ..logger import get_logger
from .base_instrument import BaseInstrument
from .daq_instrument import DAQInstrument
from .mpm_instrument import MPMInstrument
from .tsl_instrument import TSLInstrument
from .wrapper import InstrumentWrapper
from .wrapper.enumerations.connection_enums import ConnectionType, GPIBType, Terminator


class InstrumentManager:
    """Main instrument manager for device detection and connection"""

    def __init__(self):
        """Initializes the InstrumentManager."""
        self._resources = []
        self.logger = get_logger(self.__class__.__name__)
        self._instrument_wrapper = InstrumentWrapper(self.logger)
        self._instrument: BaseInstrument | None = None
        self._connected_instruments: Dict[str, BaseInstrument] = {}
        self._resources_listed: bool = False
        self.logger.info("Initializing Instrument Manager...")

    # region Private methods
    def _list_resources(self):
        """Lists all available resources."""
        if not self._resources_listed:
            self.logger.info("Listing all available resources...")
            self._resources.clear()
            try:
                self._list_gpib_resources()
                self._list_usb_resources()
                self._list_daq_resources()

                self._resources_listed = True

            except Exception as e:
                error_string = f"Error while listing resources: {e}"
                self.logger.error(error_string)
                raise Exception(error_string)

    def _list_gpib_resources(self):
        """Lists GPIB resources."""
        self.logger.info("Listing VISA GPIB resources...")

        try:
            gpib_resources = self._instrument_wrapper.get_gpib_resources()
            if gpib_resources:
                self._resources.extend(gpib_resources)

        except Exception as e:
            self.logger.error(f"Error listing VISA GPIB resources: {e}")

    def _list_usb_resources(self):
        """Lists FTDI USB resources."""
        self.logger.info("Listing FTDI USB resources...")
        try:
            usb_resources = self._instrument_wrapper.get_usb_resources()
            if usb_resources:
                self._resources.extend(usb_resources)
        except Exception as e:
            self.logger.error(f"Error listing FTDI USB resources: {e}")

    def _list_daq_resources(self):
        """Lists NI DAQ devices."""
        self.logger.info("Listing NI DAQ resources...")
        try:
            daq_devices = self._instrument_wrapper.get_daq_devices()
            if daq_devices:
                self._resources.extend(daq_devices)
        except Exception as e:
            self.logger.error(f"Error listing NI DAQ resources: {e}")

    def _list_serial_port_resources(self):
        """Lists Serial Port devices."""
        self.logger.info("Listing Serial Port resources...")
        try:
            serial_port_devices = self._instrument_wrapper.get_serial_ports()
            if serial_port_devices:
                self._resources.extend(serial_port_devices)
        except Exception as e:
            self.logger.error(f"Error listing Serial Port resources: {e}")

    def _connect(self, resource_name, terminator: Terminator = Terminator.CRLF):
        """Connects to the specified resource."""
        connection_type = None
        if resource_name in self._connected_instruments.keys():
            raise Exception(f"Resource {resource_name} already connected.")

        if "TCPIP" in resource_name.upper():
            connection_type = ConnectionType.TCPIP
            self._resources.append(resource_name)

        if len(self._resources) < 1:
            raise Exception(f"No resources available: {len(self._resources)}")

        if resource_name not in self._resources:
            self.logger.error(f"Try to connect invalid resource: {resource_name}")
            raise Exception(f"Invalid resource: {resource_name}")

        if not connection_type:
            connection_type = self._get_connection_type(resource_name)

        self._establish_connection(resource_name, connection_type, terminator)
        if not self._instrument:
            raise Exception(f"Failed to connect: {resource_name}")

        self._connected_instruments[resource_name] = self._instrument

        return self._instrument

    def _establish_connection(
        self,
        resource_name: str,
        connection_type: ConnectionType,
        terminator: Terminator,
    ):
        """Establishes a connection to the specified resource."""
        self.logger.info(
            f"Establishing connection to {resource_name} "
            f"of type {connection_type.name}..."
        )
        match connection_type:
            case ConnectionType.GPIB:
                self._gpib_connection(resource_name, terminator)
            case ConnectionType.USB:
                self._usb_connection(resource_name)
            case ConnectionType.TCPIP:
                self._tcpip_connection(resource_name, terminator)
            case ConnectionType.DEV:
                self._dev_connection(resource_name)
            case ConnectionType.NULL:
                raise Exception(f"Invalid connection type: {connection_type}")

    def _gpib_connection(self, resource_name, terminator):
        """Establishes a GPIB connection."""
        self.logger.info(f"Connecting to GPIB resource: {resource_name}")
        gpib_board, gpib_address, _ = resource_name.split("::")  # GPIB0::10::INSTR
        gpib_board = gpib_board[-1]
        self._instrument_wrapper.connect_gpib(
            self._instrument,
            int(gpib_board),
            int(gpib_address),
            GPIBType.NI4882,
            terminator,
        )

    def _usb_connection(self, resource_name):
        """Establishes a USB connection."""
        self.logger.info(f"Connecting to USB resource: {resource_name}")
        # usb_device_id = 1  # TODO: Refactor the usb device ID assignment
        raise NotImplementedError("USB connection is yet to be implemented.")

    def _tcpip_connection(self, resource_name, terminator):
        """Establishes a TCPIP connection."""
        self.logger.info(f"Connecting to TCPIP resource: {resource_name}")
        _, ip_address, port_number, _ = resource_name.split(
            "::"
        )  # TCPIP0::192.168.10.101::5000::SOCKET
        self._instrument_wrapper.connect_tcpip(
            self._instrument, str(ip_address), int(port_number), terminator
        )

    def _dev_connection(self, resource_name):
        """Establishes a connection to a NI DAQ device."""
        self.logger.info(f"Connecting to NI DAQ resource: {resource_name}")
        self._instrument_wrapper.connect_daq(self._instrument, resource_name)  # Dev1

    @staticmethod
    def _get_connection_type(resource_name):
        """Determines the connection type based on the resource name."""
        resource_name = resource_name.upper()
        if "GPIB" in resource_name:
            return ConnectionType.GPIB
        elif "TSL" in resource_name:
            return ConnectionType.USB
        elif "DEV" in resource_name:
            return ConnectionType.DEV
        return ConnectionType.NULL

    # endregion

    def list_resources(self) -> list:
        """
        List all available devices.

        :return: List of GPIB, FTDI USB & NI DAQ resources.
        """
        self._list_resources()
        resources = self._resources
        if len(resources) < 1:
            self.logger.debug(f"No resources available: {len(resources)}")
            return []

        self.logger.info(f"Found {len(resources)} resources")
        return resources

    def connect_tsl(self, resource_name: str) -> TSLInstrument | BaseInstrument:
        """Connects to a TSL instrument."""
        self._list_resources()
        if not resource_name:
            raise ValueError("Resource name cannot be empty.")
        self.logger.info(f"Connecting to TSL resource: {resource_name}")
        terminator = Terminator.CR
        self._instrument = TSLInstrument()
        return self._connect(resource_name, terminator)

    def connect_mpm(self, resource_name: str) -> MPMInstrument | BaseInstrument:
        """Connects to an MPM instrument."""
        self._list_resources()
        if not resource_name:
            raise ValueError("Resource name cannot be empty.")
        self.logger.info(f"Connecting to MPM resource: {resource_name}")
        terminator = Terminator.LF
        self._instrument = MPMInstrument()
        return self._connect(resource_name, terminator)

    def connect_daq(self, device_name: str) -> DAQInstrument | BaseInstrument:
        """Connects to a NI DAQ device."""
        self._list_resources()
        if not device_name:
            raise ValueError("Device name cannot be empty.")
        self.logger.info(f"Connecting to NI DAQ device: {device_name}")
        self._instrument = DAQInstrument()
        return self._connect(device_name)



"""
Santec Instrument DLL Wrapper.
"""

import Santec

# Santec Communication Terminator Enum class
CommunicationTerminator = Santec.CommunicationTerminator


class TSL(Santec.TSL):
    """Wrapper for the Santec TSL instrument."""

    pass


class MPM(Santec.MPM):
    """Wrapper for the Santec MPM instrument."""

    pass


class DAQ(Santec.SPU):
    """Wrapper for the Santec DAQ (SPU Class) instrument."""

    pass



# pysantec/instruments/wrapper/__init__.py

"""
Santec Instrument DLL Wrapper.
"""

from .enumerations import connection_enums, mpm_enums, tsl_enums
from .instrument_wrapper import InstrumentWrapper
from .santec_wrapper import DAQ, MPM, TSL
from .exceptions import InstrumentExceptionCode, to_instrument_exception_code

__all__ = [
    "InstrumentWrapper",
    "TSL",
    "MPM",
    "DAQ",
    "connection_enums",
    "tsl_enums",
    "mpm_enums",
    "InstrumentExceptionCode",
    "to_instrument_exception_code",
]



"""
PySantec logger module.
"""

import datetime
import logging
import os
import platform
import sys

from . import __about__

# Ensure log directory exists
LOG_DIR = "logs"
os.makedirs(LOG_DIR, exist_ok=True)

# Date and time for logging
dt = datetime.datetime.now()
dt = dt.strftime("%Y%m%d")

# Configure the logger
LOG_FILE = os.path.join(LOG_DIR, f"output_{dt}.log")

# Logging Level
LOGGING_LEVEL = logging.INFO

# Initialize root logger
root_logger = logging.getLogger()
if not root_logger.hasHandlers():
    logging.basicConfig(
        level=LOGGING_LEVEL,
        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
        handlers=[
            logging.FileHandler(LOG_FILE, mode="w"),
            # logging.StreamHandler()
        ],
    )

# Log the project version
logging.info(f"Project Version: {__about__.__version__}")


def _log_run_info():
    """Log environment information."""
    info = [
        f"Python Version: {sys.version}",
        f"Python Implementation: {platform.python_implementation()}",
        f"Architecture: {platform.architecture()[0]}",
        f"Operating System: {platform.system()} {platform.release()}",
        f"Platform ID: {platform.platform()}",
        f"Machine: {platform.machine()}",
        f"Processor: {platform.processor()}",
    ]
    for line in info:
        logging.info(line)


# Log the environment information
_log_run_info()


# Return the logger
def get_logger(name: str) -> logging.Logger:
    """Get a logger with the specified name."""
    return logging.getLogger(name)



"""
Unified wrapper for the Santec Instrument DLL.
"""

from logging import Logger
from typing import Any, List, Optional

from .enumerations.connection_enums import ConnectionType, GPIBType, Terminator
from .exceptions import InstrumentConnectionError, InstrumentOperationError
from .santec_communication_wrapper import MainCommunication
from .santec_wrapper import DAQ


class InstrumentWrapper:
    """Unified wrapper for the Santec Instrument DLL."""

    def __init__(self, logger: Logger):
        """Initialize the wrapper with logging capability.

        Args:
            logger: Logger instance for tracking operations
        """
        self.logger = logger.manager.getLogger(__class__.__name__)
        self._main_comm = MainCommunication()
        self._spu = DAQ()

    def get_usb_resources(self) -> List[str]:
        """Get available USB resources.

        Returns:
            List of USB resource identifiers
        """
        resources = list(self._main_comm.Get_USB_Resouce())
        self.logger.debug(f"USB resources found: {len(resources)}")
        return resources

    def get_gpib_resources(self) -> List[str]:
        """Get available GPIB resources.

        Returns:
            List of GPIB resource identifiers
        """
        resources = list(self._main_comm.Get_GPIB_Resources())
        self.logger.debug(f"GPIB resources found: {len(resources)}")
        return resources

    def get_serial_ports(self) -> List[str]:
        """Get available serial ports.

        Returns:
            List of serial port identifiers
        """
        resources = list(self._main_comm.Get_Serial_Port())
        self.logger.debug(f"Serial port resources found: {len(resources)}")
        return resources

    def get_daq_devices(self) -> Optional[List[str]]:
        """Get available DAQ devices.

        Returns:
            List of DAQ device identifiers or None if no devices found

        Raises:
            DeviceError: If there's an error getting DAQ devices
        """
        response = self._spu.Get_Device_ID(None)
        error_code, devices = response[0], list(response[1])

        self.logger.debug(f"DAQ devices found: {len(devices)}")

        if error_code == -11:
            self.logger.debug("No DAQ devices connected")
            return None
        if error_code != 0:
            raise InstrumentOperationError(
                "Error while getting DAQ devices", error_code
            )

        return devices

    @staticmethod
    def _create_instrument(instrument_instance: Any) -> Any:
        """Create and validate instrument instance.

        Args:
            instrument_instance: Instance to create instrument from

        Returns:
            Created instrument

        Raises:
            ConnectionError: If instrument creation fails
        """
        instrument = instrument_instance.instrument(InstrumentWrapper)
        if not instrument:
            raise InstrumentConnectionError("Could not create instrument instance")
        return instrument

    def connect_gpib(
        self,
        instrument_instance: Any,
        gpib_board: int,
        gpib_address: int,
        gpib_connect_type: GPIBType,
        terminator: Terminator,
    ) -> None:
        """Connect to instrument via GPIB.

        Args:
            instrument_instance: Instance to connect to
            gpib_board: GPIB board number
            gpib_address: GPIB address
            gpib_connect_type: Type of GPIB connection
            terminator: Message terminator type

        Raises:
            ConnectionError: If connection fails
        """
        instrument = self._create_instrument(instrument_instance)

        instrument.GPIBBoard = gpib_board
        instrument.GPIBAddress = gpib_address
        instrument.GPIBConnectType = gpib_connect_type.value
        instrument.Terminator = terminator.value

        try:
            error_code = instrument.Connect(ConnectionType.GPIB.value)
            if error_code != 0:
                raise InstrumentConnectionError(
                    f"Failed to establish GPIB connection "
                    f"with GPIB{gpib_board}::{gpib_address}",
                    error_code,
                )
        except Exception as e:
            raise InstrumentConnectionError(
                f"Error connecting to GPIB{gpib_board}::{gpib_address}",
                str(e),
            )

    def connect_tcpip(
        self,
        instrument_instance: Any,
        ip_address: str,
        port_number: int,
        terminator: Terminator,
    ) -> None:
        """Connect to instrument via TCP/IP.

        Args:
            instrument_instance: Instance to connect to
            ip_address: IP address of the instrument
            port_number: Port number
            terminator: Message terminator type

        Raises:
            ConnectionError: If connection fails
        """
        instrument = self._create_instrument(instrument_instance)

        instrument.IPAddress = ip_address
        instrument.Port = port_number
        instrument.TimeOut = 5000
        instrument.Terminator = terminator.value

        try:
            error_code = instrument.Connect(ConnectionType.TCPIP.value)
            if error_code != 0:
                raise InstrumentConnectionError(
                    f"Failed to establish LAN connection "
                    f"with {ip_address}::{port_number}",
                    error_code,
                )
        except Exception as e:
            raise InstrumentConnectionError(
                f"Error connecting to {ip_address}::{port_number}", str(e)
            )

    def connect_daq(self, instrument_instance: Any, device_name: str) -> None:
        """Connect to DAQ device.

        Args:
            instrument_instance: Instance to connect to
            device_name: Name of the DAQ device

        Raises:
            ConnectionError: If connection fails
        """
        instrument = self._create_instrument(instrument_instance)
        instrument.DeviceName = device_name

        try:
            error_code, device = instrument.Connect("")
            if error_code != 0:
                raise InstrumentConnectionError(
                    f"Failed to establish DAQ connection with {device_name}",
                    error_code,
                )
        except Exception as e:
            raise InstrumentConnectionError(
                f"Error connecting to DAQ {device_name}", str(e)
            )



# pysantec/__init__.py

"""
PySantec - Python Package for Santec Insertion Loss
and Polarization Dependent Loss Swept Test System.
"""

from .logger import get_logger
from .drivers import load_dlls

logger = get_logger(__name__)

try:
    # Initialize and Load the Santec DLLs
    setup_dlls_result = load_dlls()
    logger.info("Santec DLLs loaded successfully.")
except Exception as e:
    logger.error("Error while Santec DLLs: ", str(e))
    raise


from .instruments.instrument_manager import InstrumentManager
from .measurements.single_measurement_operation import SME


__all__ = [
    "InstrumentManager",
    "SME"
]



"""
DAQ instrument module.
"""

from ..logger import get_logger
from .base_instrument import BaseInstrument
from .wrapper import DAQ


class DAQInstrument(BaseInstrument):
    """DAQ Instrument class for handling data acquisition operations."""

    def __init__(self):
        """Initialize the DAQ instrument."""
        super().__init__()
        self._instrument = DAQ()
        self.logger = get_logger(self.__class__.__name__)
        self.logger.info("Initializing DAQ Instrument...")

    # region Properties
    @property
    def idn(self):
        """Return the IDN of the instrument."""
        self.logger.info("Retrieving IDN of the DAQ Instrument...")
        return self._instrument.DeviceName

    @property
    def is_sampling(self):
        """Check if the instrument is currently sampling."""
        self.logger.debug("Checking if the DAQ Instrument is sampling...")
        return self._instrument.IsSampling

    @property
    def logging_error_code(self):
        """Return the logging error code."""
        self.logger.debug("Retrieving logging error code...")
        return self._instrument.Logging_Errorcode

    @property
    def is_connected(self):
        """Check if the instrument is connected."""
        self.logger.debug("Checking if the DAQ Instrument is connected...")
        return self._instrument.IsConnected

    # endregion

    # region Setter & Getter Methods
    def get_devices(self):
        """Get a list of connected DAQ devices."""
        device_list = list(self._get_function("Get_Device_ID", None))
        self.logger.info(f"Connected DAQ devices: {device_list}")
        return device_list

    # Time Coefficient
    def get_time_coefficient(self):
        """Get the time coefficient of the instrument."""
        time_coefficient = self._instrument.Time_coefficient
        self.logger.debug(f"Time Coefficient: {time_coefficient}")
        return time_coefficient

    def set_time_coefficient(self, value: float):
        """Set the time coefficient of the instrument."""
        self.logger.info(f"Setting Time Coefficient to {value}.")
        self._instrument.Time_coefficient = value

    # Averaging Time
    def get_averaging_time(self):
        """Get the averaging time of the instrument."""
        averaging_time = self._instrument.AveragingTime
        self.logger.debug(f"Averaging Time: {averaging_time}")
        return averaging_time

    def set_averaging_time(self, value: float):
        """Set the averaging time of the instrument."""
        self.logger.info(f"Setting Averaging Time to {value}.")
        self._instrument.AveragingTime = value

    # F Additional Time
    def get_f_additional_time(self):
        """Get the additional time factor for the instrument."""
        f_additional_time = self._instrument.F_AdditonalTime
        self.logger.debug(f"F Additional Time: {f_additional_time}")
        return f_additional_time

    def set_f_additional_time(self, value: float):
        """Set the additional time factor for the instrument."""
        self.logger.info(f"Setting Additional Time Factor to {value}.")
        self._instrument.F_AdditonalTime = value

    # Add Time Coefficient
    def get_add_time_coefficient(self):
        """Get the additional time coefficient for the instrument."""
        add_time_coefficient = self._instrument.AddTime_coefficient
        self.logger.debug(f"Add Time Coefficient: {add_time_coefficient}")
        return add_time_coefficient

    def set_add_time_coefficient(self, value: float):
        """Set the additional time coefficient for the instrument."""
        self.logger.info(f"Setting Additional Time Coefficient to {value}.")
        self._instrument.AddTime_coefficient = value

    # Measurement Sampling Time
    def get_meas_sampling_time(self):
        """Get the measurement sampling time of the instrument."""
        meas_sampling_time = self._instrument.Meas_Sampling_time
        self.logger.debug(f"Measurement Sampling Time: {meas_sampling_time}")
        return meas_sampling_time

    def set_meas_sampling_time(self, value: float):
        """Set the measurement sampling time of the instrument."""
        self.logger.info(f"Setting Measurement Sampling Time to {value}.")
        self._instrument.Meas_Sampling_time = value

    # endregion

    # region Scan Related Methods
    def set_sampling_parameters(
        self,
        start_wavelength: float,
        stop_wavelength: float,
        speed: int,
        tsl_actual_step: float,
    ):
        """Set the sampling parameters for the instrument."""
        self.logger.info(
            f"Setting Sampling Parameters: "
            f"Start Wavelength: {start_wavelength}, "
            f"Stop Wavelength: {stop_wavelength}, "
            f"Speed: {speed}, "
            f"TSL Actual Step: {tsl_actual_step}."
        )
        self._set_function(
            "Set_Sampling_Parameter",
            start_wavelength,
            stop_wavelength,
            speed,
            tsl_actual_step,
        )

    def start_sampling(self):
        """Start the sampling process."""
        self.logger.info("Starting the sampling process.")
        self._set_function("Sampling_Start")

    def wait_for_sampling(self):
        """Wait for the sampling process to complete."""
        self.logger.info("Waiting for the sampling process to complete.")
        self._set_function("Waiting_for_sampling")

    def stop_sampling(self):
        """Stop the sampling process."""
        self.logger.info("Stopping the sampling process.")
        self._set_function("Sampling_Stop")

    def get_sampling_data(self):
        """Get the sampling data from the instrument."""
        self.logger.info("Retrieving sampling data.")

        trigger, monitor = self._get_multiple_responses("Get_Sampling_Data", None, None)

        if not trigger or not monitor:
            self.logger.error("Failed to retrieve sampling data.")
            raise ValueError("No data received from the instrument.")

        if len(trigger) != len(monitor):
            self.logger.error("Mismatch in lengths of trigger and monitor data.")
            raise ValueError("Trigger and monitor data lengths do not match.")

        self.logger.info(
            f"Retrieved {len(trigger)} trigger "
            f"and {len(monitor)} monitor data points."
        )

        return trigger, monitor

    def get_sampling_raw_data(self):
        """Get the raw sampling data from the instrument."""
        self.logger.info("Retrieving raw sampling data...")

        trigger, monitor = self._get_multiple_responses(
            "Get_Sampling_Rawdata", None, None
        )

        if not trigger or not monitor:
            self.logger.error("Failed to retrieve raw sampling data.")
            raise ValueError("No raw data received from the instrument.")

        if len(trigger) != len(monitor):
            self.logger.error("Mismatch in lengths of trigger and monitor raw data.")
            raise ValueError("Trigger and monitor raw data lengths do not match.")

        self.logger.info(
            f"Retrieved {len(trigger)} trigger "
            f"and {len(monitor)} monitor raw data points."
        )

        return trigger, monitor

    # endregion
